package pins25.phase;

import pins25.common.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * Sintaksni analizator.
 */
public class SynAn implements AutoCloseable {

    /**
     * Leksikalni analizator.
     */
    private final LexAn lexAn;

    private HashMap<AST.Node, Report.Locatable> attrLoc;

    /**
     * Ustvari nov sintaksni analizator.
     *
     * @param srcFileName Ime izvorne datoteke.
     */
    public SynAn(final String srcFileName) {
        this.lexAn = new LexAn(srcFileName);
    }

    @Override
    public void close() {
        lexAn.close();
    }

    /**
     * Prevzame leksikalni analizator od leksikalnega analizatorja in preveri, ali
     * je prave vrste.
     *
     * @param symbol Pričakovana vrsta leksikalnega simbola.
     * @return Prevzeti leksikalni simbol.
     */
    private Token check(Token.Symbol symbol) {
        final Token token = lexAn.takeToken();
        if (token.symbol() != symbol)
            throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected " + symbol + ".");
        return token;
    }

    /**
     * Opravi sintaksno analizo.
     */
    public AST.Node parse(HashMap<AST.Node, Report.Locatable> attrLoc) {
        this.attrLoc = attrLoc;
        final AST.Nodes<AST.MainDef> defs = parseProgram();
        if (lexAn.peekToken().symbol() != Token.Symbol.EOF)
            Report.warning(lexAn.peekToken(), "Unexpected text '" + lexAn.peekToken().lexeme() + "...' at the end of the program.");
        return defs;
    }

    private AST.Nodes<AST.MainDef> parseProgram() {
        // program -> definition restdefs
        List<AST.MainDef> defs = new ArrayList<>();

        defs.add(parseDefinition());
        parseRestDefinitions(defs);

        return new AST.Nodes<>(defs);
    }

    private void parseRestDefinitions(List<AST.MainDef> defs) {
        if (lexAn.peekToken().symbol() != Token.Symbol.EOF) {
            // restdefs -> definition restdefs
            defs.add(parseDefinition());
            parseRestDefinitions(defs);
        }
        // restdefs -> ε
    }

    private AST.MainDef parseDefinition() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case FUN: {
                // definition -> FUN IDENTIFIER LPAREN parameters RPAREN funcassign
                Token fun = check(Token.Symbol.FUN);
                Token id = check(Token.Symbol.IDENTIFIER);
                check(Token.Symbol.LPAREN);
                List<AST.ParDef> params = parseParameters();
                Token rParen = check(Token.Symbol.RPAREN);
                List<AST.Stmt> stmts = parseFunctionAssignment();

                AST.FunDef funDef = new AST.FunDef(id.lexeme(), params, stmts);
                attrLoc.put(funDef, new Report.Location(fun, stmts.isEmpty() ? rParen : attrLoc.get(stmts.getLast())));
                return funDef;
            }

            case VAR: {
                // definition -> VAR IDENTIFIER ASSIGN initializers
                Token var = check(Token.Symbol.VAR);
                Token id = check(Token.Symbol.IDENTIFIER);
                Token assign = check(Token.Symbol.ASSIGN);
                List<AST.Init> inits = parseInitializers();

                AST.VarDef varDef = new AST.VarDef(id.lexeme(), inits);
                attrLoc.put(varDef, inits.isEmpty() ? new Report.Location(var, assign) : new Report.Location(var, attrLoc.get(inits.getLast())));
                return varDef;
            }

            default:
                throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected definition.");
        }
    }

    private List<AST.Stmt> parseFunctionAssignment() {
        Token token = lexAn.peekToken();
        if (token.symbol() == Token.Symbol.ASSIGN) {
            // funcassign -> ASSIGN statements
            check(Token.Symbol.ASSIGN);
            return parseStatements();
        }

        // funcassign -> ε

        return new ArrayList<>();
    }

    private List<AST.ParDef> parseParameters() {
        List<AST.ParDef> defs = new ArrayList<>();
        Token token = lexAn.peekToken();
        if (token.symbol() == Token.Symbol.IDENTIFIER) {
            // parameters -> IDENTIFIER restparams
            Token id = check(Token.Symbol.IDENTIFIER);
            AST.ParDef parDef = new AST.ParDef(id.lexeme());
            attrLoc.put(parDef, id);
            defs.add(parDef);
            parseRestParameters(defs);
        }

        // parameters -> ε

        return defs;
    }

    private void parseRestParameters(List<AST.ParDef> defs) {
        Token token = lexAn.peekToken();
        if (token.symbol() == Token.Symbol.COMMA) {
            // restparams -> COMMA IDENTIFIER restparams
            check(Token.Symbol.COMMA);
            Token id = check(Token.Symbol.IDENTIFIER);
            AST.ParDef parDef = new AST.ParDef(id.lexeme());
            attrLoc.put(parDef, id);
            defs.add(parDef);
            parseRestParameters(defs);
        }
        // restparams -> ε
    }

    private List<AST.Stmt> parseStatements() {
        // statements -> statement reststmts

        List<AST.Stmt> stmts = new ArrayList<>();

        stmts.add(parseStatement());
        parseRestStatements(stmts);

        return stmts;
    }

    private AST.Expr parseExpressionAssign() {
        if (lexAn.peekToken().symbol() == Token.Symbol.ASSIGN) {
            // exprassign -> ASSIGN expression
            check(Token.Symbol.ASSIGN);
            return parseExpression();
        }

        // exprassign -> ε
        return null;
    }

    private List<AST.Stmt> parseElseStatements() {
        if (lexAn.peekToken().symbol() == Token.Symbol.ELSE) {
            // elsestmt -> ELSE statements
            check(Token.Symbol.ELSE);
            return parseStatements();
        }

        // elsestmt -> ε
        return new ArrayList<>();
    }

    private void parseRestStatements(List<AST.Stmt> stmts) {
        if (lexAn.peekToken().symbol() == Token.Symbol.COMMA) {
            // reststmts -> COMMA statement reststmts
            check(Token.Symbol.COMMA);
            stmts.add(parseStatement());
            parseRestStatements(stmts);
        }
        // reststmts -> ε
    }

    private AST.Stmt parseStatement() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case IF: {
                // statement -> if expression then statements elsestmt end
                Token ifT = check(Token.Symbol.IF);
                AST.Expr cond = parseExpression();
                check(Token.Symbol.THEN);
                List<AST.Stmt> thenStmts = parseStatements();
                List<AST.Stmt> elseStmts = parseElseStatements();
                Token end = check(Token.Symbol.END);

                AST.IfStmt ifStmt = new AST.IfStmt(cond, thenStmts, elseStmts);
                attrLoc.put(ifStmt, new Report.Location(ifT, end));
                return ifStmt;
            }

            case WHILE: {
                // statement -> while expression do statements end
                Token whileT = check(Token.Symbol.WHILE);
                AST.Expr cond = parseExpression();
                check(Token.Symbol.DO);
                List<AST.Stmt> stmts = parseStatements();
                Token end = check(Token.Symbol.END);

                AST.WhileStmt whileStmt = new AST.WhileStmt(cond, stmts);
                attrLoc.put(whileStmt, new Report.Location(whileT, end));
                return whileStmt;
            }

            case LET: {
                // statement -> let definition reststmtdefs in statements end
                Token let = check(Token.Symbol.LET);
                List<AST.MainDef> defs = new ArrayList<>();
                defs.add(parseDefinition());
                parseRestStatementDefinitions(defs);
                check(Token.Symbol.IN);
                List<AST.Stmt> stmts = parseStatements();
                Token end = check(Token.Symbol.END);

                AST.LetStmt letStmt = new AST.LetStmt(defs, stmts);
                attrLoc.put(letStmt, new Report.Location(let, end));
                return letStmt;
            }

            case IDENTIFIER:
            case LPAREN:
            case ADD:
            case SUB:
            case NOT:
            case PTR:
            case INTCONST:
            case CHARCONST:
            case STRINGCONST:
                // statement -> expression exprassign
                AST.Expr expr = parseExpression();
                AST.Expr srcExpr = parseExpressionAssign();
                if (srcExpr != null) {
                    AST.AssignStmt assignStmt = new AST.AssignStmt(expr, srcExpr);
                    attrLoc.put(assignStmt, new Report.Location(attrLoc.get(expr), attrLoc.get(srcExpr)));
                    return assignStmt;
                }
                AST.ExprStmt exprStmt = new AST.ExprStmt(expr);
                attrLoc.put(exprStmt, attrLoc.get(expr));
                return exprStmt;

            default:
                throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected statement.");
        }
    }

    private void parseRestStatementDefinitions(List<AST.MainDef> defs) {
        switch (lexAn.peekToken().symbol()) {
            case FUN:
            case VAR:
                // reststmtdefs -> definition reststmtdefs
                defs.add(parseDefinition());
                parseRestStatementDefinitions(defs);
                break;

            case IN:
                // reststmtdefs -> ε
                break;
        }
    }

    private List<AST.Init> parseInitializers() {
        List<AST.Init> inits = new ArrayList<>();

        switch (lexAn.peekToken().symbol()) {
            case INTCONST:
            case CHARCONST:
            case STRINGCONST:
                // initializers -> initializer restinits
                inits.add(parseInitializer());
                parseRestInitializers(inits);
                break;

            default:
                // initializers -> ε
                break;
        }

        return inits;
    }

    private AST.Init parseInitializer() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case INTCONST: {
                // initializer -> INTCONST intconstmult
                Token value = check(Token.Symbol.INTCONST);
                AST.AtomExpr intConst = new AST.AtomExpr(AST.AtomExpr.Type.INTCONST, value.lexeme());
                attrLoc.put(intConst, value);
                AST.AtomExpr multConst = parseIntegerConstantMultiplier();
                if (multConst != null) {
                    AST.Init init = new AST.Init(intConst, multConst);
                    attrLoc.put(init, new Report.Location(value, attrLoc.get(multConst)));
                    return init;
                }
                AST.Init init = new AST.Init(new AST.AtomExpr(AST.AtomExpr.Type.INTCONST, "1"), intConst);
                attrLoc.put(init, value);
                return init;
            }

            case CHARCONST: {
                // initializer -> CHARCONST
                Token value = check(Token.Symbol.CHARCONST);
                AST.AtomExpr charConst = new AST.AtomExpr(AST.AtomExpr.Type.CHRCONST, value.lexeme());
                attrLoc.put(charConst, value);
                AST.Init init = new AST.Init(new AST.AtomExpr(AST.AtomExpr.Type.INTCONST, "1"), charConst);
                attrLoc.put(init, value);
                return init;
            }

            case STRINGCONST: {
                // initializer -> STRINGCONST
                Token value = check(Token.Symbol.STRINGCONST);
                AST.AtomExpr strConst = new AST.AtomExpr(AST.AtomExpr.Type.STRCONST, value.lexeme());
                attrLoc.put(strConst, value);
                AST.Init init = new AST.Init(new AST.AtomExpr(AST.AtomExpr.Type.INTCONST, "1"), strConst);
                attrLoc.put(init, value);
                return init;
            }

            default:
                throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected constant.");
        }
    }

    private void parseRestInitializers(List<AST.Init> inits) {
        Token token = lexAn.peekToken();
        if (token.symbol() == Token.Symbol.COMMA) {
            // restinits -> COMMA initializer restinits
            check(Token.Symbol.COMMA);
            inits.add(parseInitializer());
            parseRestInitializers(inits);
        }
        // restinits -> ε
    }

    private AST.AtomExpr parseIntegerConstantMultiplier() {
        Token token = lexAn.peekToken();
        if (token.symbol() == Token.Symbol.MUL) {
            // intconstmult -> MUL const
            check(Token.Symbol.MUL);
            return parseConstant();
        }
        // intconstmult -> ε
        return null;
    }

    private AST.AtomExpr parseConstant() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case INTCONST: {
                // const -> INTCONST
                Token value = check(Token.Symbol.INTCONST);
                AST.AtomExpr atomExpr = new AST.AtomExpr(AST.AtomExpr.Type.INTCONST, value.lexeme());
                attrLoc.put(atomExpr, value);
                return atomExpr;
            }
            case CHARCONST: {
                // const -> CHARCONST
                Token value = check(Token.Symbol.CHARCONST);
                AST.AtomExpr atomExpr = new AST.AtomExpr(AST.AtomExpr.Type.CHRCONST, value.lexeme());
                attrLoc.put(atomExpr, value);
                return atomExpr;
            }
            case STRINGCONST: {
                // const -> STRINGCONST
                Token value = check(Token.Symbol.STRINGCONST);
                AST.AtomExpr atomExpr = new AST.AtomExpr(AST.AtomExpr.Type.STRCONST, value.lexeme());
                attrLoc.put(atomExpr, value);
                return atomExpr;
            }

            default:
                throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected constant.");
        }
    }

    private void checkExpression() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case IDENTIFIER:
            case LPAREN:
            case ADD:
            case SUB:
            case NOT:
            case PTR:
            case INTCONST:
            case CHARCONST:
            case STRINGCONST:
                return;

            default:
                throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected expression.");
        }
    }

    private AST.Expr parseExpression() {
        checkExpression();

        // expression -> conjexpr restdisj
        AST.Expr left = parseConjunctionExpression();
        return parseRestDisjunctions(left);
    }

    private AST.Expr parseRestDisjunctions(AST.Expr left) {
        if (lexAn.peekToken().symbol() == Token.Symbol.OR) {
            // restdisj -> OR conjexpr restdisj
            check(Token.Symbol.OR);
            AST.Expr right = parseConjunctionExpression();
            AST.Expr binExpr = new AST.BinExpr(AST.BinExpr.Oper.OR, left, right);
            attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
            return parseRestDisjunctions(binExpr);
        } else {
            // restdisj -> ε
            return left;
        }
    }

    private AST.Expr parseConjunctionExpression() {
        checkExpression();

        // conjexpr -> cmpexpr restconj
        AST.Expr left = parseComparisonExpression();
        return parseRestConjunctions(left);
    }

    private AST.Expr parseRestConjunctions(AST.Expr left) {
        if (lexAn.peekToken().symbol() == Token.Symbol.AND) {
            // restconj -> AND cmpexpr restconj
            check(Token.Symbol.AND);
            AST.Expr right = parseComparisonExpression();
            AST.Expr binExpr = new AST.BinExpr(AST.BinExpr.Oper.AND, left, right);
            attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
            return parseRestConjunctions(binExpr);
        } else {
            // restconj -> ε
            return left;
        }
    }

    private AST.Expr parseComparisonExpression() {
        checkExpression();

        // cmpexpr -> addexpr restcmp
        AST.Expr left = parseAdditionExpression();
        return parseRestComparisons(left);
    }

    private AST.Expr parseRestComparisons(AST.Expr left) {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case EQU: {
                // restcmp -> EQU addexpr
                check(Token.Symbol.EQU);
                AST.Expr right = parseAdditionExpression();
                AST.BinExpr binExpr = new AST.BinExpr(AST.BinExpr.Oper.EQU, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return binExpr;
            }
            case NEQ: {
                // restcmp -> NEQ addexpr
                check(Token.Symbol.NEQ);
                AST.Expr right = parseAdditionExpression();
                AST.BinExpr binExpr = new AST.BinExpr(AST.BinExpr.Oper.NEQ, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return binExpr;
            }
            case LTH: {
                // restcmp -> LTH addexpr
                check(Token.Symbol.LTH);
                AST.Expr right = parseAdditionExpression();
                AST.BinExpr binExpr = new AST.BinExpr(AST.BinExpr.Oper.LTH, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return binExpr;
            }
            case GTH: {
                // restcmp -> GTH addexpr
                check(Token.Symbol.GTH);
                AST.Expr right = parseAdditionExpression();
                AST.BinExpr binExpr = new AST.BinExpr(AST.BinExpr.Oper.GTH, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return binExpr;
            }
            case LEQ: {
                // restcmp -> LEQ addexpr
                check(Token.Symbol.LEQ);
                AST.Expr right = parseAdditionExpression();
                AST.BinExpr binExpr = new AST.BinExpr(AST.BinExpr.Oper.LEQ, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return binExpr;
            }
            case GEQ: {
                // restcmp -> GEQ addexpr
                check(Token.Symbol.GEQ);
                AST.Expr right = parseAdditionExpression();
                AST.BinExpr binExpr = new AST.BinExpr(AST.BinExpr.Oper.GEQ, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return binExpr;
            }

            default:
                // restcmp -> ε
                return left;
        }
    }

    private AST.Expr parseAdditionExpression() {
        checkExpression();

        // addexpr -> multexpr restadd
        AST.Expr left = parseMultiplicationExpression();
        return parseRestAdditionExpressions(left);
    }

    private AST.Expr parseRestAdditionExpressions(AST.Expr left) {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case ADD: {
                // restadd -> ADD multexpr restadd
                check(Token.Symbol.ADD);
                AST.Expr right = parseMultiplicationExpression();
                AST.Expr binExpr = new AST.BinExpr(AST.BinExpr.Oper.ADD, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return parseRestAdditionExpressions(binExpr);
            }
            case SUB: {
                // restadd -> SUB multexpr restadd
                check(Token.Symbol.SUB);
                AST.Expr right = parseMultiplicationExpression();
                AST.Expr binExpr = new AST.BinExpr(AST.BinExpr.Oper.SUB, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return parseRestAdditionExpressions(binExpr);
            }

            default:
                // restadd -> ε
                return left;
        }
    }

    private AST.Expr parseMultiplicationExpression() {
        checkExpression();

        // multexpr -> prefixexpr restmult
        AST.Expr left = parsePrefixExpression();
        return parseRestMultiplicationExpressions(left);
    }

    private AST.Expr parseRestMultiplicationExpressions(AST.Expr left) {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case MUL: {
                // restmult -> MUL prefixexpr restmult
                check(Token.Symbol.MUL);
                AST.Expr right = parsePrefixExpression();
                AST.Expr binExpr = new AST.BinExpr(AST.BinExpr.Oper.MUL, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return parseRestMultiplicationExpressions(binExpr);
            }
            case DIV: {
                // restmult -> DIV prefixexpr restmult
                check(Token.Symbol.DIV);
                AST.Expr right = parsePrefixExpression();
                AST.Expr binExpr = new AST.BinExpr(AST.BinExpr.Oper.DIV, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return parseRestMultiplicationExpressions(binExpr);
            }
            case MOD: {
                // restmult -> MOD prefixexpr restmult
                check(Token.Symbol.MOD);
                AST.Expr right = parsePrefixExpression();
                AST.Expr binExpr = new AST.BinExpr(AST.BinExpr.Oper.MOD, left, right);
                attrLoc.put(binExpr, new Report.Location(attrLoc.get(left), attrLoc.get(right)));
                return parseRestMultiplicationExpressions(binExpr);
            }

            default:
                // restmult -> ε
                return left;
        }
    }

    private AST.Expr parsePrefixExpression() {
        checkExpression();

        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case ADD: {
                // prefixexpr -> ADD prefixexpr
                Token add = check(Token.Symbol.ADD);
                AST.Expr prefixExpr = parsePrefixExpression();
                AST.UnExpr unExpr = new AST.UnExpr(AST.UnExpr.Oper.ADD, prefixExpr);
                attrLoc.put(unExpr, new Report.Location(add, attrLoc.get(prefixExpr)));
                return unExpr;
            }
            case SUB: {
                // prefixexpr -> SUB prefixexpr
                Token sub = check(Token.Symbol.SUB);
                AST.Expr prefixExpr = parsePrefixExpression();
                AST.UnExpr unExpr = new AST.UnExpr(AST.UnExpr.Oper.SUB, prefixExpr);
                attrLoc.put(unExpr, new Report.Location(sub, attrLoc.get(prefixExpr)));
                return unExpr;
            }
            case NOT: {
                // prefixexpr -> NOT prefixexpr
                Token not = check(Token.Symbol.NOT);
                AST.Expr prefixExpr = parsePrefixExpression();
                AST.UnExpr unExpr = new AST.UnExpr(AST.UnExpr.Oper.NOT, prefixExpr);
                attrLoc.put(unExpr, new Report.Location(not, attrLoc.get(prefixExpr)));
                return unExpr;
            }
            case PTR: {
                // prefixexpr -> PTR prefixexpr
                Token ptr = check(Token.Symbol.PTR);
                AST.Expr prefixExpr = parsePrefixExpression();
                AST.UnExpr unExpr = new AST.UnExpr(AST.UnExpr.Oper.MEMADDR, prefixExpr);
                attrLoc.put(unExpr, new Report.Location(ptr, attrLoc.get(prefixExpr)));
                return unExpr;
            }

            default:
                // prefixexpr -> postfixexpr
                return parsePostfixExpression();
        }
    }

    private AST.Expr parsePostfixExpression() {
        checkExpression();

        // postfixexpr -> primary postfixop
        AST.Expr inner = parsePrimary();
        return parsePostfixOperator(inner);
    }

    private AST.Expr parsePostfixOperator(AST.Expr inner) {
        if (lexAn.peekToken().symbol() == Token.Symbol.PTR) {
            // postfixop -> PTR postfixop
            Token ptr = check(Token.Symbol.PTR);
            AST.UnExpr unExpr = new AST.UnExpr(AST.UnExpr.Oper.VALUEAT, inner);
            attrLoc.put(unExpr, new Report.Location(attrLoc.get(inner), ptr));
            return parsePostfixOperator(unExpr);
        }
        // postfixop -> ε
        return inner;
    }

    private AST.Expr parsePrimary() {
        Token token = lexAn.peekToken();
        switch (token.symbol()) {
            case IDENTIFIER:
                // primary -> IDENTIFIER exprargs
                Token id = check(Token.Symbol.IDENTIFIER);
                if (lexAn.peekToken().symbol() == Token.Symbol.LPAREN) {
                    // exprargs -> LPAREN arguments RPAREN
                    check(Token.Symbol.LPAREN);
                    List<AST.Expr> args = parseArguments();
                    Token rParen = check(Token.Symbol.RPAREN);

                    AST.CallExpr callExpr = new AST.CallExpr(id.lexeme(), args);
                    attrLoc.put(callExpr, new Report.Location(id, rParen));
                    return callExpr;
                }

                AST.VarExpr varExpr = new AST.VarExpr(id.lexeme());
                attrLoc.put(varExpr, id);
                return varExpr;

            case LPAREN:
                // primary -> LPAREN expression RPAREN
                Token lParen = check(Token.Symbol.LPAREN);
                AST.Expr expr = parseExpression();
                Token rParen = check(Token.Symbol.RPAREN);
                attrLoc.put(expr, new Report.Location(lParen, rParen));
                return expr;

            case INTCONST:
            case CHARCONST:
            case STRINGCONST:
                // primary -> const
                return parseConstant();

            default:
                throw new Report.Error(token, "Unexpected symbol '" + token.lexeme() + "', expected identifier, constant or '('.");
        }
    }

    private List<AST.Expr> parseArguments() {
        List<AST.Expr> args = new ArrayList<>();

        Token token = lexAn.peekToken();
        if (token.symbol() == Token.Symbol.RPAREN) {
            // arguments -> ε
            return args;
        }
        // arguments -> expression restargs
        args.add(parseExpression());
        parseRestArguments(args);

        return args;
    }

    private void parseRestArguments(List<AST.Expr> args) {
        if (lexAn.peekToken().symbol() == Token.Symbol.COMMA) {
            // restargs -> COMMA restargs
            check(Token.Symbol.COMMA);
            args.add(parseExpression());
            parseRestArguments(args);
        }
    }


    // --- ZAGON ---

    /**
     * Zagon sintaksnega analizatorja kot samostojnega programa.
     *
     * @param cmdLineArgs Argumenti v ukazni vrstici.
     */
    public static void main(final String[] cmdLineArgs) {
        System.out.println("This is PINS'25 compiler (syntax analysis):");

        try {
            if (cmdLineArgs.length == 0) throw new Report.Error("No source file specified in the command line.");
            if (cmdLineArgs.length > 1) Report.warning("Unused arguments in the command line.");

            try (SynAn synAn = new SynAn(cmdLineArgs[0])) {
                synAn.parse(new HashMap<>());
            }

            // Upajmo, da kdaj pridemo to te tocke.
            // A zavedajmo se sledecega:
            // 1. Prevod je zaradi napak v programu lahko napacen :-o
            // 2. Izvorni program se zdalec ni tisto, kar je programer hotel, da bi bil ;-)
            Report.info("Done.");
        } catch (Report.Error error) {
            // Izpis opisa napake.
            System.err.println(error.getMessage());
            System.exit(1);
        }
    }

}
